from packaging.version import Version
from unittest import SkipTest

import numpy as np
import pandas as pd
import hvplot.pandas  # noqa
import pytest

from hvplot import hvPlotTabular
from hvplot.tests.util import makeDataFrame

try:
    import polars as pl
    import hvplot.polars  # noqa

    skip_polar = False
except ImportError:

    class pl:
        DataFrame = None
        LazyFrame = None
        Series = None

    skip_polar = True

try:
    import dask.dataframe as dd
    import hvplot.dask  # noqa
except ImportError:
    dd = None


TYPES = {t for t in dir(hvPlotTabular) if not t.startswith('_') and t != 'explorer'}
FRAME_TYPES = TYPES - {'bivariate', 'heatmap', 'hexbin', 'labels', 'vectorfield'}
SERIES_TYPES = FRAME_TYPES - {'points', 'polygons', 'ohlc', 'paths'}
frame_kinds = pytest.mark.parametrize('kind', sorted(FRAME_TYPES))
series_kinds = pytest.mark.parametrize('kind', sorted(SERIES_TYPES))

y_combinations = pytest.mark.parametrize(
    'y',
    (
        ['A', 'B', 'C', 'D'],
        ('A', 'B', 'C', 'D'),
        {'A', 'B', 'C', 'D'},
        np.array(['A', 'B', 'C', 'D']),
        pd.Index(['A', 'B', 'C', 'D']),
        pd.Series(['A', 'B', 'C', 'D']),
    ),
    ids=lambda x: type(x).__name__,
)


@frame_kinds
@y_combinations
def test_dataframe_pandas(kind, y):
    df = makeDataFrame()
    df.hvplot(y=y, kind=kind)


@series_kinds
def test_series_pandas(kind):
    ser = pd.Series(np.random.rand(10), name='A')
    ser.hvplot(kind=kind)


@pytest.mark.skipif(dd is None, reason='dask not installed')
@frame_kinds
@y_combinations
def test_dataframe_dask(kind, y):
    df = dd.from_pandas(makeDataFrame(), npartitions=2)
    assert isinstance(df, dd.DataFrame)
    df.hvplot(y=y, kind=kind)


@pytest.mark.skipif(dd is None, reason='dask not installed')
@series_kinds
def test_series_dask(kind):
    ser = dd.from_pandas(pd.Series(np.random.rand(10), name='A'), npartitions=2)
    assert isinstance(ser, dd.Series)
    ser.hvplot(kind=kind)


@pytest.mark.skipif(skip_polar, reason='polars not installed')
@pytest.mark.parametrize('cast', (pl.DataFrame, pl.LazyFrame))
@frame_kinds
@y_combinations
def test_dataframe_polars(kind, y, cast):
    df = cast(makeDataFrame())
    assert isinstance(df, cast)
    df.hvplot(y=y, kind=kind)


@pytest.mark.skipif(skip_polar, reason='polars not installed')
@series_kinds
def test_series_polars(kind):
    ser = pl.Series(values=np.random.rand(10), name='A')
    assert isinstance(ser, pl.Series)
    ser.hvplot(kind=kind)


@pytest.mark.skipif(skip_polar, reason='polars not installed')
@series_kinds
def test_series_polars_downstream(kind):
    if Version(pl.__version__) < Version('0.20.3'):
        raise SkipTest('plot namespace in Polars introduced in 0.20.3')
    ser = pl.Series(values=np.random.rand(10), name='A')
    assert isinstance(ser, pl.Series)
    ser.plot(kind=kind)



import sys

from .interactive import Interactive


class DaskInteractive(Interactive):
    @classmethod
    def applies(cls, obj):
        if 'dask.dataframe' in sys.modules:
            import dask.dataframe as dd

            return isinstance(obj, (dd.Series, dd.DataFrame))
        return False

    def compute(self):
        self._method = 'compute'
        return self.__call__()


def patch(name='hvplot', interactive='interactive', extension='bokeh', logo=False):
    from . import hvPlotTabular, post_patch

    try:
        import dask.dataframe as dd
    except ImportError:
        raise ImportError('Could not patch plotting API onto dask. Dask could not be imported.')
    _patch_plot = lambda self: hvPlotTabular(self)  # noqa: E731
    _patch_plot.__doc__ = hvPlotTabular.__call__.__doc__
    plot_prop = property(_patch_plot)
    setattr(dd.DataFrame, name, plot_prop)
    setattr(dd.Series, name, plot_prop)

    _patch_interactive = lambda self: DaskInteractive(self)  # noqa: E731
    _patch_interactive.__doc__ = DaskInteractive.__call__.__doc__
    interactive_prop = property(_patch_interactive)
    setattr(dd.DataFrame, interactive, interactive_prop)
    setattr(dd.Series, interactive, interactive_prop)

    post_patch(extension, logo)


patch()



from packaging.version import Version

from . import hvPlot, post_patch


def patch(name='hvplot', extension='bokeh', logo=False):
    try:
        import intake
    except ImportError:
        raise ImportError(
            'Could not patch plotting API onto intake. intake could not be imported.'
        )

    _patch_plot = lambda self: hvPlot(self)  # noqa: E731
    _patch_plot.__doc__ = hvPlot.__call__.__doc__
    patch_property = property(_patch_plot)
    setattr(intake.source.base.DataSource, name, patch_property)
    post_patch(extension, logo)


try:
    import intake.plotting  # noqa

    patch()
except Exception:
    import intake

    if Version(intake.__version__) <= Version('0.1.5'):
        patch()
        patch(name='plot')
    else:
        post_patch(extension='bokeh', logo=False)



import dask

optional_markers = {
    'geo': {
        'help': 'Run the tests that require GeoViews',
        'marker-descr': 'Geo test marker',
        'skip-reason': 'Test only runs with the --geo option.',
    },
}


def pytest_addoption(parser):
    for marker, info in optional_markers.items():
        parser.addoption(
            '--{}'.format(marker), action='store_true', default=False, help=info['help']
        )


def pytest_configure(config):
    for marker, info in optional_markers.items():
        config.addinivalue_line('markers', '{}: {}'.format(marker, info['marker-descr']))


def pytest_collection_modifyitems(config, items):
    skipped, selected = [], []
    markers = [m for m in optional_markers if config.getoption(f'--{m}')]
    empty = not markers
    for item in items:
        if empty and any(m in item.keywords for m in optional_markers):
            skipped.append(item)
        elif empty:
            selected.append(item)
        elif not empty and any(m in item.keywords for m in markers):
            selected.append(item)
        else:
            skipped.append(item)

    config.hook.pytest_deselected(items=skipped)
    items[:] = selected


# From Dask 2024.3.0 they now use `dask_expr` by default
# https://github.com/dask/dask/issues/10995
dask.config.set({'dataframe.query-planning': False})
